"""Helper functions for data manipulation."""
import concurrent.futures
import itertools
import logging
from itertools import product
from multiprocessing import cpu_count

import numpy as np
import torch
from tqdm import tqdm
from transformers import BertTokenizer, BertModel


def compute_possible_kmers(kmer_size: int = 6, alphabet: str = "ACGT") -> np.array:
    """
    Compute all kmer possibilities based on given alphabet.

    Parameters
    ----------
    kmer_size : int, default=6
        K-mer size.
    alphabet : str, default='ACGT'
        The alphabet used to compute k-mers.

    Returns
    -------
    kmers : np.array
        Numpy array containing all possible k-mers.
    """
    kmers = ["".join(c) for c in product(alphabet, repeat=kmer_size)]
    return np.array(kmers)


logger = logging.getLogger()


def configure_device():
    """
    Configure device to run computation.

    Returns
    -------
    device: torch.device
        Device to compute, i.e., CPU or GPU.
    """
    if torch.cuda.is_available():
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")
        logger.warning("ProtBERT: CPU mode enabled (slower)")
    return device


def load_bert(device: torch.device):
    """
    Load BERT model.

    Parameters
    ----------
    device: torch.device
        Device to compute, i.e., CPU or GPU.


    Returns
    -------
    tokenizer: BertTokenizer
        BERT tokenizer.
    model: BertModel
        BERT encoder model.
    """
    tokenizer = BertTokenizer.from_pretrained(
        "zhihan1996/DNABERT-2-117M", trust_remote_code=True, do_lower_case=False
    )
    model = BertModel.from_pretrained(
        "zhihan1996/DNABERT-2-117M", trust_remote_code=True
    ).to(device)
    return tokenizer, model


def get_bert_embedding(
    sequence: str,
    len_seq_limit: int,
    tokenizer: BertTokenizer,
    device: torch.device,
    model: BertModel,
):
    """
    Collect last hidden state embedding vector from pre-trained ProtBERT Model.

    Parameters
    ----------
    sequence: str
        Protein sequence, e.g., AAABBB from FASTA file.
    len_seq_limit: int
        Maximum sequence length, i.e., number of letters for truncation.
    tokenizer: BertTokenizer
        BERT tokenizer.
    device: torch.device
        Device to compute, i.e., CPU or GPU.
    model: BertModel
        BERT encoder model.

    Returns
    -------
    output_hidden: np.ndarray
        Last hidden state embedding vector for input sequence.
    """
    sequence_w_spaces = " ".join(list(sequence))
    encoded_input = tokenizer(
        sequence_w_spaces,
        truncation=True,
        max_length=len_seq_limit,
        padding="max_length",
        return_tensors="pt",
    ).to(device)
    output = model(**encoded_input)
    output_hidden = output["last_hidden_state"][:, 0][0].detach().cpu().numpy()
    return output_hidden


def compute_features(sequences: list) -> np.array:
    """
    Compute features using DNA BERT.

    Parameters
    ----------
    sequences : list
        List containing all sequences.

    Returns
    -------
    features : np.array
        Numpy array containing features for all sequences.
    """
    device = configure_device()
    tokenizer, model = load_bert(device)
    embeddings = []
    for seq in tqdm(sequences):
        embeddings.append(
            get_bert_embedding(
                sequence=seq,
                len_seq_limit=1200,  # TODO: make this a user parameter
                tokenizer=tokenizer,
                device=device,
                model=model,
            )
        )
    return np.array(embeddings)


def compute_kmer_frequency(sequence: str, kmers: list) -> np.array:
    """
    Compute kmer frequencies for a given sequence.

    Parameters
    ----------
    sequence : str
        Sequence to compute k-mer frequency.
    kmers : list
        List containing possible k-mers to compute frequency (anything else is not counted).

    Returns
    -------
    frequency : np.array
        Numpy array containing frequency for all possible k-mers.
    """
    dictionary = {}
    for kmer in kmers:
        dictionary[kmer] = 0
    for i in range(len(sequence) - (len(kmers[0]) - 1)):
        key = sequence[i : i + len(kmers[0])]
        if key in dictionary:
            dictionary[key] = dictionary[key] + 1
    frequency = []
    for kmer in kmers:
        frequency.append(dictionary[kmer])
    return np.array(frequency)


def grouper(maximum_elements: int, items: list) -> np.array:
    """
    Return groups with a maximum of n elements for a given list.

    Parameters
    ----------
    maximum_elements : int
        Maximum number of elements per group.
    items : list
        List containing items to divide in groups.

    Returns
    -------
    groups : np.array
        Numpy array containing groups of items.
    """
    groups = []
    subgroup = []
    for item in items:
        if len(subgroup) < maximum_elements:
            subgroup.append(item)
        else:
            groups.append(subgroup)
            subgroup = [item]
    if len(subgroup) > 0:
        groups.append(subgroup)
    return np.array(groups, dtype=object)


def compute_group_frequency(sequences_and_kmers: tuple) -> np.array:
    """
    Compute k-mer frequency for a group of sequences.

    Parameters
    ----------
    sequences_and_kmers : tuple
        Tuple containing a list of sequences and k-mers.

    Returns
    -------
    frequencies : np.array
        Numpy array containing frequencies for a group of sequences.
    """
    frequencies = []
    sequences, kmers = sequences_and_kmers
    for sequence in sequences:
        frequencies.append(compute_kmer_frequency(sequence, kmers))
    return np.array(frequencies)


def compute_frequencies(
    sequences: list, kmers: list, threads: int = cpu_count(), batch_size: int = 100
) -> np.array:
    """
    Compute k-mer frequency for all sequences.

    Parameters
    ----------
    sequences : list
        List containing all sequences.
    kmers : list
        List containing all possible k-mers.
    threads : int, default='all CPUs'
        Number of threads to compute in parallel.
    batch_size : int, default=100
        Size of each batch to run in parallel.

    Returns
    -------
    frequencies : np.array
        Numpy array containing frequencies for all sequences.
    """
    sequences = [s.decode("utf-8") for s in sequences]
    executor = concurrent.futures.ProcessPoolExecutor(threads)
    futures = [
        executor.submit(compute_group_frequency, (group, kmers))
        for group in grouper(batch_size, sequences)
    ]
    concurrent.futures.wait(futures)
    frequencies = [f.result() for f in futures]
    frequencies = list(itertools.chain(*frequencies))
    return np.array(frequencies)


def extract_qiime2_ranks(taxonomy: str) -> np.array:
    """
    Split taxonomy by ranks.

    Parameters
    ----------
    taxonomy : str
        String with concatenated taxonomy.

    Returns
    -------
    ranks : np.array
        Numpy array where each item is a taxonomic rank.
    """
    ranks = taxonomy.split(";")
    return np.array(ranks)


def extract_qiime2_taxonomy(taxonomy: list) -> np.array:
    """
    Extract taxonomy from a list of strings.

    Parameters
    ----------
    taxonomy : list
        List String with concatenated taxonomy.

    Returns
    -------
    extracted_taxonomy : np.array
        Numpy array where each item is another array with taxonomic ranks.
    """
    extracted_taxonomy = []
    for record in taxonomy:
        extracted_taxonomy.append(extract_qiime2_ranks(record))
    return np.array(extracted_taxonomy)


def _extract_reads(reads) -> zip:
    """
    Extract IDs and sequences from DNAIterator.

    Parameters
    ----------
    reads : DNAIterator
        Iterator containing sequences.

    Returns
    -------
    _ : zip
        Zip object containing tuples with ID and sequence.
    """
    return zip(*[(r.metadata["id"], r._string) for r in reads])


def convert_taxonomy_to_qiime2(predictions: np.array) -> list:
    """
    Convert predictions made by HiClass to QIIME2 taxonomy format.

    Parameters
    ----------
    predictions : np.array
        Predictions made by HiClass.

    Returns
    -------
    taxonomy : list
        List containing converted taxonomy.
    """
    taxonomy = []
    for prediction in predictions:
        tax = ""
        for i in range(len(prediction) - 1):
            tax = tax + prediction[i] + ";"
        tax = tax + prediction[-1]
        taxonomy.append(tax)
    return taxonomy


# TODO: replace with binary search if speed becomes an issue
def search(key: str, items: list) -> int:
    """
    Search for a key index in a list.

    Parameters
    ----------
    key : str
        Key to search for.
    items : list
        List containing items to search.

    Returns
    -------
    index : int
        Index of the item in the list.
    """
    for i in range(len(items)):
        if key == items[i]:
            return i


def compute_confidence(
    classification: np.array,
    classes: np.array,
    predict_proba: np.array,
    threshold: float,
) -> tuple:
    """
    Compute confidence score for predictions and removes ranks if their confidence is lower than threshold.

    Parameters
    ----------
    classification : np.array
        Taxonomic predictions made by previous classifier.
    classes : np.array
        Classes in each taxonomic rank.
    predict_proba : np.array
        Prediction probability computed by the filter.
    threshold : float
        Minimum confidence score to keep taxonomic ranks

    Returns
    -------
    predictions, confidence : tuple
        Predictions with ranks above minimum threshold and confidence score for the lowest rank available.
    """
    predictions = []
    confidence = []
    for row in range(len(classification)):
        predictions.append([])
        confidence.append(-1)
        for rank in reversed(range(len(classification[row]))):
            if confidence[row] == -1:
                index = search(classification[row][rank], classes[rank])
                conf = predict_proba[rank][row][index]
                if conf >= threshold:
                    confidence[row] = conf
            if confidence[row] >= threshold:
                predictions[row].insert(0, classification[row][rank])
            else:
                predictions[row].insert(0, "")
    return predictions, confidence



"""QIIME2 public functions for HiTaC."""
from multiprocessing import cpu_count

import pandas as pd
import skbio
from hiclass import LocalClassifierPerParentNode
from q2_types.feature_data import (
    DNAFASTAFormat,
    DNAIterator,
    FeatureData,
    Sequence,
    Taxonomy,
)
from qiime2.plugin import Float, Int
from sklearn.linear_model import LogisticRegression

from ._hierarchical_taxonomic_classifier import HierarchicalTaxonomicClassifier
from ._utils import (
    _extract_reads,
    compute_confidence,
    compute_frequencies,
    compute_possible_kmers,
    convert_taxonomy_to_qiime2,
    extract_qiime2_taxonomy,
    compute_features,
)
from .filter import Filter
from .plugin_setup import citations, plugin


def fit(
    reference_reads: DNAIterator,
    reference_taxonomy: pd.Series,
    kmer: int = 6,
    threads: int = cpu_count(),
) -> LocalClassifierPerParentNode:
    """
    Fit HiTaC's classifier.

    Parameters
    ----------
    reference_reads : DNAIterator
        Reference reads.
    reference_taxonomy : pd.Series
        Reference taxonomy.
    kmer : int, default=6
        K-mer size.
    threads : int, default='All CPUs'
        Number of threads for parallel training.

    Returns
    -------
    hierarchical_classifier : LocalClassifierPerParentNode
        Local hierarchical classifier based on the taxonomic hierarchy.
    """
    _, training_sequences = _extract_reads(reference_reads)
    X_train = compute_features(
        training_sequences
    )  # TODO: add threads and gpus parameters
    Y_train = extract_qiime2_taxonomy(reference_taxonomy)
    logistic_regression = LogisticRegression(
        solver="liblinear",
        multi_class="auto",
        class_weight="balanced",
        max_iter=10000,
        verbose=0,
        n_jobs=1,
    )
    hierarchical_classifier = LocalClassifierPerParentNode(
        local_classifier=logistic_regression, n_jobs=threads
    )
    hierarchical_classifier.fit(X_train, Y_train)
    return hierarchical_classifier


plugin.methods.register_function(
    function=fit,
    inputs={
        "reference_reads": FeatureData[Sequence],
        "reference_taxonomy": FeatureData[Taxonomy],
    },
    parameters={"kmer": Int, "threads": Int},
    parameter_descriptions={
        "kmer": "K-mer size.",
        "threads": "Number of threads for parallel training",
    },
    outputs=[("classifier", HierarchicalTaxonomicClassifier)],
    name="Train HiTaC's hierarchical classifier",
    description="Train HiTaC's hierarchical classifier",
    citations=[citations["miranda2020hitac"]],
)


def classify(
    reads: DNAFASTAFormat,
    classifier: LocalClassifierPerParentNode,
    kmer: int = 6,
    threads: int = cpu_count(),
) -> pd.DataFrame:
    """
    Classify sequences with HiTaC.

    Parameters
    ----------
    reads : DNAFASTAFormat
        Reads to classify.
    classifier : LocalClassifierPerParentNode
        Pre-fitted hierarchical classifier.
    kmer : int, default=6
        K-mer size.
    threads : int, default='All CPUs'
        Number of threads for parallel classification.

    Returns
    -------
    classification : pd.DataFrame
        DataFrame containing the taxonomies assigned to each sequence.
    """
    # transform reads to DNAIterator
    reads = DNAIterator(skbio.read(str(reads), format="fasta", constructor=skbio.DNA))
    seq_ids, test_sequences = _extract_reads(reads)
    X_test = compute_features(test_sequences)  # TODO: add threads and gpu parameters
    predictions = classifier.predict(X_test)
    taxonomy = convert_taxonomy_to_qiime2(predictions)
    confidence = [-1] * len(seq_ids)
    result = pd.DataFrame(
        {"Taxon": taxonomy, "Confidence": confidence},
        index=seq_ids,
        columns=["Taxon", "Confidence"],
    )
    result.index.name = "Feature ID"
    return result


plugin.methods.register_function(
    function=classify,
    inputs={
        "reads": FeatureData[Sequence],
        "classifier": HierarchicalTaxonomicClassifier,
    },
    input_descriptions={
        "reads": "The feature data to be classified.",
        "classifier": "The hierarchical taxonomic classifier for classifying the reads.",
    },
    parameters={"kmer": Int, "threads": Int},
    parameter_descriptions={
        "kmer": "K-mer size.",
        "threads": "Number of threads for parallel classification",
    },
    outputs=[("classification", FeatureData[Taxonomy])],
    name="Hierarchical classification with HiTaC's pre-fitted model",
    description="Classify reads by taxon using a fitted hierarchical classifier.",
    citations=[citations["miranda2020hitac"]],
)


def fit_filter(
    reference_reads: DNAIterator,
    reference_taxonomy: pd.Series,
    kmer: int = 6,
    threads: int = cpu_count(),
) -> Filter:
    """
    Fit HiTaC's filter.

    Parameters
    ----------
    reference_reads : DNAIterator
        Reference reads.
    reference_taxonomy : pd.Series
        Reference taxonomy.
    kmer : int, default=6
        K-mer size.
    threads : int, default='All CPUs'
        Number of threads for parallel training.

    Returns
    -------
    hierarchical_classifier : Filter
        Local hierarchical filter based on the taxonomic hierarchy.
    """
    kmers = compute_possible_kmers(kmer)
    _, training_sequences = _extract_reads(reference_reads)
    X_train = compute_frequencies(training_sequences, kmers, threads)
    Y_train = extract_qiime2_taxonomy(reference_taxonomy)
    logistic_regression = LogisticRegression(
        solver="liblinear",
        multi_class="auto",
        class_weight="balanced",
        max_iter=10000,
        verbose=0,
        n_jobs=1,
    )
    hierarchical_classifier = Filter(
        local_classifier=logistic_regression, n_jobs=threads
    )
    hierarchical_classifier.fit(X_train, Y_train)
    return hierarchical_classifier


plugin.methods.register_function(
    function=fit_filter,
    inputs={
        "reference_reads": FeatureData[Sequence],
        "reference_taxonomy": FeatureData[Taxonomy],
    },
    parameters={"kmer": Int, "threads": Int},
    parameter_descriptions={
        "kmer": "K-mer size.",
        "threads": "Number of threads for parallel training",
    },
    outputs=[("filter", HierarchicalTaxonomicClassifier)],
    name="Train HiTaC's hierarchical filter",
    description="Train HiTaC's hierarchical filter",
    citations=[citations["miranda2020hitac"]],
)


def filter(
    reads: DNAFASTAFormat,
    filter: Filter,
    classification: pd.DataFrame,
    threshold: float = 0.7,
    kmer: int = 6,
    threads: int = cpu_count(),
) -> pd.DataFrame:
    """
    Filter sequences with HiTaC.

    Parameters
    ----------
    reads : DNAFASTAFormat
        Reads to filter.
    filter : Filter
        Pre-fitted hierarchical filter.
    classification : pd.DataFrame
        Predictions made by HiTaC's classifier.
    threshold : float, default=0.7
        Confidence threshold for limiting taxonomic depth. Set to 0 to compute confidence score but not apply it to limit the taxonomic depth of the assignments.
    kmer : int, default=6
        K-mer size.
    threads : int, default='All CPUs'
        Number of threads for parallel filtering.

    Returns
    -------
    filtered_classification : pd.DataFrame
        DataFrame containing the taxonomies assigned to each sequence and the prediction probability for the lowest taxonomic rank.
    """
    kmers = compute_possible_kmers(kmer)
    # transform reads to DNAIterator
    reads = DNAIterator(skbio.read(str(reads), format="fasta", constructor=skbio.DNA))
    seq_ids, test_sequences = _extract_reads(reads)
    X_test = compute_frequencies(test_sequences, kmers, threads)
    predict_proba = filter.predict_proba(X_test)
    classes = filter.classes_
    classification = extract_qiime2_taxonomy(classification["Taxon"])
    predictions, confidence = compute_confidence(
        classification, classes, predict_proba, threshold
    )
    taxonomy = convert_taxonomy_to_qiime2(predictions)
    result = pd.DataFrame(
        {"Taxon": taxonomy, "Confidence": confidence},
        index=seq_ids,
        columns=["Taxon", "Confidence"],
    )
    result.index.name = "Feature ID"
    return result


plugin.methods.register_function(
    function=filter,
    inputs={
        "reads": FeatureData[Sequence],
        "filter": HierarchicalTaxonomicClassifier,
        "classification": FeatureData[Taxonomy],
    },
    input_descriptions={
        "reads": "The feature data to be filtered.",
        "filter": "The hierarchical taxonomic filter for filtering the reads.",
        "classification": "The predictions made by HiTaC",
    },
    parameters={"threshold": Float, "kmer": Int, "threads": Int},
    parameter_descriptions={
        "threshold": "Confidence threshold for limiting taxonomic depth. Set to 0 to compute confidence score but not apply it to limit the taxonomic depth of the assignments.",
        "kmer": "K-mer size.",
        "threads": "Number of threads for parallel filtering",
    },
    outputs=[("filtered_classification", FeatureData[Taxonomy])],
    name="Hierarchical classification filtering with HiTaC's pre-fitted model",
    description="Filter reads using a fitted hierarchical filter.",
    citations=[citations["miranda2020hitac"]],
)



import itertools
import os

import pandas as pd
import skbio
from hitac import classifier
from q2_types.feature_data import DNAIterator

fixtures_loc = os.path.join(os.path.dirname(__file__), "fixtures")


class TestClassifier:
    def test_classify_1(self):
        reference_reads = DNAIterator(
            skbio.read(
                os.path.join(fixtures_loc, "classify_1_reference_reads.fasta"),
                format="fasta",
                constructor=skbio.DNA,
            )
        )
        reference_taxonomy = list(
            itertools.chain(
                *pd.read_csv(
                    os.path.join(fixtures_loc, "classify_1_reference_taxonomy.tsv"),
                    sep="\t",
                    usecols=["Taxon"],
                ).values.tolist()
            )
        )
        query_reads = os.path.join(fixtures_loc, "classify_1_query_reads.fasta")
        ground_truth = list(
            itertools.chain(
                *pd.read_csv(
                    os.path.join(fixtures_loc, "classify_1_ground_truth.tsv"),
                    sep="\t",
                    usecols=["Taxon"],
                ).values.tolist()
            )
        )
        ground_truth.sort()
        hierarchical_classifier = classifier.fit(reference_reads, reference_taxonomy)
        predictions = classifier.classify(query_reads, hierarchical_classifier)[
            "Taxon"
        ].values.tolist()
        predictions.sort()
        assert len(ground_truth) == len(predictions)
        assert all([a == b for a, b in zip(predictions, ground_truth)])

    def test_filter_1(self):
        reference_reads_1 = DNAIterator(
            skbio.read(
                os.path.join(fixtures_loc, "classify_1_reference_reads.fasta"),
                format="fasta",
                constructor=skbio.DNA,
            )
        )
        reference_reads_2 = DNAIterator(
            skbio.read(
                os.path.join(fixtures_loc, "classify_1_reference_reads.fasta"),
                format="fasta",
                constructor=skbio.DNA,
            )
        )
        reference_taxonomy = list(
            itertools.chain(
                *pd.read_csv(
                    os.path.join(fixtures_loc, "classify_1_reference_taxonomy.tsv"),
                    sep="\t",
                    usecols=["Taxon"],
                ).values.tolist()
            )
        )
        query_reads = os.path.join(fixtures_loc, "classify_1_query_reads.fasta")
        ground_truth = list(
            itertools.chain(
                *pd.read_csv(
                    os.path.join(fixtures_loc, "classify_1_ground_truth.tsv"),
                    sep="\t",
                    usecols=["Taxon"],
                ).values.tolist()
            )
        )
        ground_truth.sort()
        hierarchical_classifier = classifier.fit(reference_reads_1, reference_taxonomy)
        classification = classifier.classify(query_reads, hierarchical_classifier)
        hierarchical_filter = classifier.fit_filter(
            reference_reads_2, reference_taxonomy
        )
        predictions = classifier.filter(
            query_reads, hierarchical_filter, classification
        )["Taxon"].values.tolist()
        predictions.sort()
        assert len(ground_truth) == len(predictions)
        assert all([a == b for a, b in zip(predictions, ground_truth)])



"""QIIME2 private functions for HiTaC."""
import json
import os
import tarfile

import joblib
import qiime2.plugin
import qiime2.plugin.model as model
import sklearn
from hiclass import LocalClassifierPerParentNode

from .filter import Filter
from .plugin_setup import plugin

# Semantic Types
HierarchicalTaxonomicClassifier = qiime2.plugin.SemanticType(
    "HierarchicalTaxonomicClassifier"
)


# Formats
class PickleFormat(model.BinaryFileFormat):
    """Class for PickleFormat."""

    def sniff(self):
        """
        Check if file is TAR.

        Returns
        -------
        _ : bool
            True if file is TAR compressed, otherwise false.
        """
        return tarfile.is_tarfile(str(self))


# https://github.com/qiime2/q2-types/issues/49
class JSONFormat(model.TextFileFormat):
    """Class for JSONFormat."""

    def sniff(self):
        """
        Check JSON file is valid.

        Returns
        -------
        _ : bool
            True if file is JSON file is valid, otherwise false.
        """
        with self.open() as fh:
            try:
                json.load(fh)
                return True
            except json.JSONDecodeError:
                pass
        return False


class HierarchicalTaxonomicClassifierDirFmt(model.DirectoryFormat):
    """Class for HierarchicalTaxonomicClassifierDirFmt."""

    preprocess_params = model.File("preprocess_params.json", format=JSONFormat)
    sklearn_pipeline = model.File("sklearn_pipeline.tar", format=PickleFormat)


class HierarchicalTaxonomicClassiferTemporaryPickleDirFmt(model.DirectoryFormat):
    """Class for HierarchicalTaxonomicClassiferTemporaryPickleDirFmt."""

    version_info = model.File("sklearn_version.json", format=JSONFormat)
    sklearn_pipeline = model.File("sklearn_pipeline.tar", format=PickleFormat)


# Transformers
@plugin.register_transformer
def _1(
    dirfmt: HierarchicalTaxonomicClassiferTemporaryPickleDirFmt,
) -> LocalClassifierPerParentNode:
    sklearn_version = dirfmt.version_info.view(dict)["sklearn-version"]
    if sklearn_version != sklearn.__version__:
        raise ValueError(
            "The scikit-learn version (%s) used to generate this"
            " artifact does not match the current version"
            " of scikit-learn installed (%s). Please retrain your"
            " classifier for your current deployment to prevent"
            " data-corruption errors." % (sklearn_version, sklearn.__version__)
        )

    sklearn_pipeline = dirfmt.sklearn_pipeline.view(PickleFormat)

    with tarfile.open(str(sklearn_pipeline)) as tar:
        tmpdir = model.DirectoryFormat()
        dirname = str(tmpdir)
        tar.extractall(dirname)
        pipeline = joblib.load(os.path.join(dirname, "sklearn_pipeline.pkl"))
        for fn in tar.getnames():
            os.unlink(os.path.join(dirname, fn))

    return pipeline


@plugin.register_transformer
def _2(
    data: LocalClassifierPerParentNode,
) -> HierarchicalTaxonomicClassiferTemporaryPickleDirFmt:
    sklearn_pipeline = PickleFormat()
    with tarfile.open(str(sklearn_pipeline), "w") as tar:
        tmpdir = model.DirectoryFormat()
        pf = os.path.join(str(tmpdir), "sklearn_pipeline.pkl")
        for fn in joblib.dump(data, pf):
            tar.add(fn, os.path.basename(fn))
            os.unlink(fn)

    dirfmt = HierarchicalTaxonomicClassiferTemporaryPickleDirFmt()
    dirfmt.version_info.write_data({"sklearn-version": sklearn.__version__}, dict)
    dirfmt.sklearn_pipeline.write_data(sklearn_pipeline, PickleFormat)

    return dirfmt


@plugin.register_transformer
def _3(dirfmt: HierarchicalTaxonomicClassifierDirFmt) -> LocalClassifierPerParentNode:
    raise ValueError(
        "The scikit-learn version could not be determined for"
        " this artifact, please retrain your classifier for your"
        " current deployment to prevent data-corruption errors."
    )


@plugin.register_transformer
def _4(fmt: JSONFormat) -> dict:
    with fmt.open() as fh:
        return json.load(fh)


@plugin.register_transformer
def _5(data: dict) -> JSONFormat:
    result = JSONFormat()
    with result.open() as fh:
        json.dump(data, fh)
    return result


# Transformers
@plugin.register_transformer
def _6(
    dirfmt: HierarchicalTaxonomicClassiferTemporaryPickleDirFmt,
) -> Filter:
    sklearn_version = dirfmt.version_info.view(dict)["sklearn-version"]
    if sklearn_version != sklearn.__version__:
        raise ValueError(
            "The scikit-learn version (%s) used to generate this"
            " artifact does not match the current version"
            " of scikit-learn installed (%s). Please retrain your"
            " classifier for your current deployment to prevent"
            " data-corruption errors." % (sklearn_version, sklearn.__version__)
        )

    sklearn_pipeline = dirfmt.sklearn_pipeline.view(PickleFormat)

    with tarfile.open(str(sklearn_pipeline)) as tar:
        tmpdir = model.DirectoryFormat()
        dirname = str(tmpdir)
        tar.extractall(dirname)
        pipeline = joblib.load(os.path.join(dirname, "sklearn_pipeline.pkl"))
        for fn in tar.getnames():
            os.unlink(os.path.join(dirname, fn))

    return pipeline


@plugin.register_transformer
def _7(
    data: Filter,
) -> HierarchicalTaxonomicClassiferTemporaryPickleDirFmt:
    sklearn_pipeline = PickleFormat()
    with tarfile.open(str(sklearn_pipeline), "w") as tar:
        tmpdir = model.DirectoryFormat()
        pf = os.path.join(str(tmpdir), "sklearn_pipeline.pkl")
        for fn in joblib.dump(data, pf):
            tar.add(fn, os.path.basename(fn))
            os.unlink(fn)

    dirfmt = HierarchicalTaxonomicClassiferTemporaryPickleDirFmt()
    dirfmt.version_info.write_data({"sklearn-version": sklearn.__version__}, dict)
    dirfmt.sklearn_pipeline.write_data(sklearn_pipeline, PickleFormat)

    return dirfmt


@plugin.register_transformer
def _8(dirfmt: HierarchicalTaxonomicClassifierDirFmt) -> Filter:
    raise ValueError(
        "The scikit-learn version could not be determined for"
        " this artifact, please retrain your classifier for your"
        " current deployment to prevent data-corruption errors."
    )


# Registrations
plugin.register_semantic_types(HierarchicalTaxonomicClassifier)
plugin.register_formats(
    HierarchicalTaxonomicClassifierDirFmt,
    HierarchicalTaxonomicClassiferTemporaryPickleDirFmt,
)
plugin.register_semantic_type_to_format(
    HierarchicalTaxonomicClassifier,
    artifact_format=HierarchicalTaxonomicClassiferTemporaryPickleDirFmt,
)



import numpy as np
from hitac import _utils
from q2_types.feature_data import DNAIterator
from skbio.sequence._dna import DNA


class TestUtils:
    def test_compute_possible_kmers_1(self):
        ground_truth = np.array(["A", "C", "G", "T"])
        results = _utils.compute_possible_kmers(1)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_compute_possible_kmers_2(self):
        ground_truth = np.array(
            [
                "AA",
                "AC",
                "AG",
                "AT",
                "CA",
                "CC",
                "CG",
                "CT",
                "GA",
                "GC",
                "GG",
                "GT",
                "TA",
                "TC",
                "TG",
                "TT",
            ]
        )
        results = _utils.compute_possible_kmers(2)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_frequency_1(self):
        ground_truth = np.array([1, 1, 2, 1])
        kmers = _utils.compute_possible_kmers(1)
        results = _utils.compute_kmer_frequency("ATCGG", kmers)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_frequency_2(self):
        ground_truth = np.array([0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 0])
        kmers = _utils.compute_possible_kmers(2)
        results = _utils.compute_kmer_frequency("ATCGG", kmers)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_grouper_1(self):
        elements = 5
        items = [1, 2, 3, 4, 5]
        ground_truth = np.array([items])
        results = _utils.grouper(elements, items)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_grouper_2(self):
        elements = 3
        items = [1, 2, 3, 4, 5]
        ground_truth = np.array([[1, 2, 3], [4, 5]])
        results = _utils.grouper(elements, items)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_compute_frequencies_1(self):
        sequences = (b"CCAACC", b"CGGGCC")
        kmers = _utils.compute_possible_kmers(1)
        threads = 1
        ground_truth = np.array([[2, 4, 0, 0], [0, 3, 3, 0]])
        results = _utils.compute_frequencies(sequences, kmers, threads)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_computeFrequencies_2(self):
        sequences = (b"CCAACC", b"CGGGCC")
        kmers = _utils.compute_possible_kmers(2)
        threads = 2
        ground_truth = np.array(
            [
                [1, 1, 0, 0, 1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 1, 0, 0, 1, 2, 0, 0, 0, 0, 0],
            ]
        )
        results = _utils.compute_frequencies(sequences, kmers, threads)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_ranks_1(self):
        taxonomy = (
            "d__Fungi; p__Ascomycota; c__Orbiliomycetes; "
            "o__Orbiliales; f__Orbiliaceae; g__Dactylellina; "
            "s__Dactylellina_cionopaga"
        )
        ground_truth = np.array(
            [
                "d__Fungi",
                " p__Ascomycota",
                " c__Orbiliomycetes",
                " o__Orbiliales",
                " f__Orbiliaceae",
                " g__Dactylellina",
                " s__Dactylellina_cionopaga",
            ]
        )
        results = _utils.extract_qiime2_ranks(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_ranks_2(self):
        taxonomy = (
            "d__Fungi; p__Basidiomycota; "
            "c__Agaricomycetes; o__Agaricales; "
            "f__Amanitaceae; g__Limacella; s__Limacella_illinita"
        )
        ground_truth = np.array(
            [
                "d__Fungi",
                " p__Basidiomycota",
                " c__Agaricomycetes",
                " o__Agaricales",
                " f__Amanitaceae",
                " g__Limacella",
                " s__Limacella_illinita",
            ]
        )
        results = _utils.extract_qiime2_ranks(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_ranks_3(self):
        taxonomy = (
            "d__Fungi; p__Ascomycota; c__Sordariomycetes; "
            "o__Sordariales; f__Chaetomiaceae; g__Myceliophthora"
        )
        ground_truth = np.array(
            [
                "d__Fungi",
                " p__Ascomycota",
                " c__Sordariomycetes",
                " o__Sordariales",
                " f__Chaetomiaceae",
                " g__Myceliophthora",
            ]
        )
        results = _utils.extract_qiime2_ranks(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_ranks_4(self):
        taxonomy = (
            "d__Fungi; p__Ascomycota; c__Sordariomycetes; "
            "o__Sordariales; f__Chaetomiaceae"
        )
        ground_truth = np.array(
            [
                "d__Fungi",
                " p__Ascomycota",
                " c__Sordariomycetes",
                " o__Sordariales",
                " f__Chaetomiaceae",
            ]
        )
        results = _utils.extract_qiime2_ranks(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_taxonomy_1(self):
        taxonomy_1 = (
            "d__Fungi; p__Basidiomycota; "
            "c__Agaricomycetes; o__Agaricales; "
            "f__Inocybaceae; g__Inocybe"
        )
        taxonomy_2 = (
            "d__Fungi; p__Basidiomycota; "
            "c__Agaricomycetes; o__Hymenochaetales; "
            "f__Hymenochaetaceae; g__Phellinus"
        )
        taxonomy_3 = (
            "d__Fungi; p__Basidiomycota; "
            "c__Pucciniomycetes; o__Pucciniales; "
            "f__Melampsoraceae; g__Melampsora"
        )
        taxonomy = np.array([taxonomy_1, taxonomy_2, taxonomy_3])
        ground_truth = np.array(
            [
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Agaricales",
                    " f__Inocybaceae",
                    " g__Inocybe",
                ],
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Hymenochaetales",
                    " f__Hymenochaetaceae",
                    " g__Phellinus",
                ],
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Pucciniomycetes",
                    " o__Pucciniales",
                    " f__Melampsoraceae",
                    " g__Melampsora",
                ],
            ]
        )
        results = _utils.extract_qiime2_taxonomy(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_taxonomy_2(self):
        taxonomy_1 = (
            "d__Fungi; p__Basidiomycota; "
            "c__Agaricomycetes; o__Agaricales; "
            "f__Mycenaceae; g__Panellus"
        )
        taxonomy_2 = (
            "d__Fungi; p__Basidiomycota; "
            "c__Agaricomycetes; o__Atheliales; "
            "f__Atheliaceae; g__Piloderma"
        )
        taxonomy_3 = (
            "d__Fungi; p__Ascomycota; c__Dothideomycetes; "
            "o__Pleosporales; f__Didymellaceae; g__Didymella"
        )
        taxonomy = np.array([taxonomy_1, taxonomy_2, taxonomy_3])
        ground_truth = np.array(
            [
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Agaricales",
                    " f__Mycenaceae",
                    " g__Panellus",
                ],
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Atheliales",
                    " f__Atheliaceae",
                    " g__Piloderma",
                ],
                [
                    "d__Fungi",
                    " p__Ascomycota",
                    " c__Dothideomycetes",
                    " o__Pleosporales",
                    " f__Didymellaceae",
                    " g__Didymella",
                ],
            ]
        )
        results = _utils.extract_qiime2_taxonomy(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_qiime2_taxonomy_3(self):
        taxonomy_1 = (
            "d__Fungi; p__Ascomycota; c__Lecanoromycetes; "
            "o__Lecanorales; f__Parmeliaceae; g__Usnea; s__Usnea_ceratina"
        )
        taxonomy_2 = (
            "d__Fungi; p__Basidiomycota; c__Agaricomycetes; "
            "o__Agaricales; f__Marasmiaceae; g__Gymnopus; s__Gymnopus_parvulus"
        )
        taxonomy_3 = (
            "d__Fungi; p__Basidiomycota; c__Agaricomycetes; "
            "o__Agaricales; f__Psathyrellaceae; g__Parasola; "
            "s__Parasola_leiocephala"
        )
        taxonomy = np.array([taxonomy_1, taxonomy_2, taxonomy_3])
        ground_truth = np.array(
            [
                [
                    "d__Fungi",
                    " p__Ascomycota",
                    " c__Lecanoromycetes",
                    " o__Lecanorales",
                    " f__Parmeliaceae",
                    " g__Usnea",
                    " s__Usnea_ceratina",
                ],
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Agaricales",
                    " f__Marasmiaceae",
                    " g__Gymnopus",
                    " s__Gymnopus_parvulus",
                ],
                [
                    "d__Fungi",
                    " p__Basidiomycota",
                    " c__Agaricomycetes",
                    " o__Agaricales",
                    " f__Psathyrellaceae",
                    " g__Parasola",
                    " s__Parasola_leiocephala",
                ],
            ]
        )
        results = _utils.extract_qiime2_taxonomy(taxonomy)
        assert ground_truth.shape == results.shape
        assert np.array_equal(ground_truth, results)

    def test_extract_reads_1(self):
        s1 = DNA("ACGT", metadata={"id": "s1"})
        s2 = DNA("TGCA", metadata={"id": "s2"})
        it = DNAIterator([s1, s2])
        ground_truth = [
            (s1.metadata["id"], s1._string),
            (s2.metadata["id"], s2._string),
        ]
        results = list(zip(*_utils._extract_reads(it)))
        assert len(ground_truth) == len(results)
        assert all([a == b for a, b in zip(results, ground_truth)])

    def test_convert_taxonomy_to_qiime2_1(self):
        taxonomy_1 = [
            "d__Fungi",
            " p__Ascomycota",
            " c__Lecanoromycetes",
            " o__Lecanorales",
            " f__Parmeliaceae",
            " g__Parmotrema",
        ]
        taxonomy_2 = [
            "d__Fungi",
            " p__Ascomycota",
            " c__Eurotiomycetes",
            " o__Eurotiales",
            " f__Trichocomaceae",
            " g__Penicillium",
        ]
        taxonomy_3 = [
            "d__Fungi",
            " p__Ascomycota",
            " c__Dothideomycetes",
            " o__Botryosphaeriales",
            " f__Botryosphaeriaceae",
            " g__Botryosphaeria",
        ]
        taxonomy = np.array([taxonomy_1, taxonomy_2, taxonomy_3])
        ground_truth = [
            "d__Fungi; p__Ascomycota; c__Lecanoromycetes; o__Lecanorales; f__Parmeliaceae; g__Parmotrema",
            "d__Fungi; p__Ascomycota; c__Eurotiomycetes; o__Eurotiales; f__Trichocomaceae; g__Penicillium",
            "d__Fungi; p__Ascomycota; c__Dothideomycetes; o__Botryosphaeriales; f__Botryosphaeriaceae; g__Botryosphaeria",
        ]
        results = _utils.convert_taxonomy_to_qiime2(taxonomy)
        assert len(ground_truth) == len(results)
        assert all([a == b for a, b in zip(results, ground_truth)])

    def test_search_1(self):
        key = "d__Fungi"
        taxonomy = [
            "d__Fungi",
            " p__Ascomycota",
            " c__Lecanoromycetes",
            " o__Lecanorales",
            " f__Parmeliaceae",
            " g__Parmotrema",
        ]
        ground_truth = 0
        result = _utils.search(key, taxonomy)
        assert ground_truth == result

    def test_search_2(self):
        key = " o__Lecanorales"
        taxonomy = [
            "d__Fungi",
            " p__Ascomycota",
            " c__Lecanoromycetes",
            " o__Lecanorales",
            " f__Parmeliaceae",
            " g__Parmotrema",
        ]
        ground_truth = 3
        result = _utils.search(key, taxonomy)
        assert ground_truth == result



import numpy as np
from sklearn.ensemble import RandomForestClassifier

from hitac.filter import Filter


def test_predict_proba_default():
    X = np.array([[1, 2], [2, 3], [3, 4], [4, 5]])
    Y = np.array([[1, 2, 3], [1, 4, 5], [1, 6, 7], [1, 2, 8]])
    filter = Filter(
        local_classifier=RandomForestClassifier(random_state=0),
        replace_classifiers=False,
        n_jobs=2,
    )
    filter.fit(X, Y)
    result = filter.predict_proba(X)
    ground_truth = [
        np.array([[1.0], [1.0], [1.0], [1.0]]),
        np.array(
            [
                [0.64, 0.32, 0.04],
                [0.27, 0.69, 0.04],
                [0.07, 0.3, 0.63],
                [0.75, 0.08, 0.17],
            ]
        ),
        np.array(
            [
                [0.64, 0.0, 0.32, 0.04],
                [0.27, 0.0, 0.69, 0.04],
                [0.01, 0.06, 0.3, 0.63],
                [0.01, 0.74, 0.08, 0.17],
            ]
        ),
    ]
    for i in range(max(len(result), len(ground_truth))):
        assert np.array_equal(result[i], ground_truth[i])



"""Extends the LocalClassifierPerLevel."""
from typing import List

import numpy as np
from hiclass import LocalClassifierPerLevel


class Filter(LocalClassifierPerLevel):
    """Add the predict_proba method and classes_ attribute."""

    def fit(self, X, y):
        """
        Fit a local classifier per level.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.
        y : array-like of shape (n_samples, n_levels)
            The target values, i.e., hierarchical class labels for classification.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        super().fit(X, y)
        self.classes_ = [
            self._remove_separator(classifier.classes_)
            for classifier in self.local_classifiers_
        ]
        return self

    def predict_proba(self, X: np.ndarray) -> List:
        """
        Compute prediction probabilities.

        Parameters
        ----------
        X : np.array of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        probabilities : np.ndarray of shape (n_levels, n_samples)
            Prediction probabilities for all classes in each hierarchical level.
        """
        probabilities = []
        for classifier in self.local_classifiers_:
            probabilities.append(classifier.predict_proba(X))
        return probabilities

    def _remove_separator(self, classes: np.ndarray) -> np.ndarray:
        return np.array([value.split(self.separator_)[-1] for value in classes])



"""QIIME2 plugin setup."""
from qiime2.plugin import Citations, Plugin

import hitac

citations = Citations.load("citations.bib", package="hitac")

plugin = Plugin(
    name="hitac",
    version=hitac.__version__,
    website="https://gitlab.com/dacs-hpi/hitac",
    package="hitac",
    description=(
        "This QIIME 2 plugin wraps HiTaC for hierarchical taxonomic classification."
    ),
    short_description="Plugin for hierarchical taxonomic classification.",
    citations=[citations["miranda2020hitac"]],
)



"""Init module for the library."""
import importlib

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

importlib.import_module("hitac._utils")
if importlib.util.find_spec("q2_types") is not None:
    importlib.import_module("hitac.classifier")
